#include "postgres.h"
#include "fmgr.h"
#include "executor/spi.h"
#include "commands/trigger.h"

/* This structure is used to maintain a session variable. It is  */
/* a doubly-linked list. Therefore, if there are a large number */
/* of session variables, this code performs poorly (linearly). */

typedef struct SessionVariable {
  char *key;
  int keylen;
  char *value;
  struct SessionVariable *parent;
  struct SessionVariable *child;
} SessionVariable;

/* A static variable used to maintain our map. Remember these? */
/* I didn't. */

static SessionVariable *session_variables = NULL;

/* This helper function is used to initialize the key or */
/* value member of a SessionVariable with the text object */
/* handed to this routine. It will replace the value of */
/* an attribute if it already exists. */

static void setVariableAttribute(SessionVariable *v, text *t, int key) {
  
  char **attrib;
  int32 len;

  /* Determine which attribute to set */

  if (key) {
    attrib = &v->key;
  } else {
    attrib = &v->value;
  }
  if (*attrib != 0) free(*attrib);

  /* Perform a deep copy */

  len = VARSIZE(t) - VARHDRSZ;
  (*attrib) = malloc(len + 1);
  memcpy((*attrib), VARDATA(t), len);
  (*attrib)[len] = 0;

  /* If we're setting the key, cache its length */

  if (key) v->keylen = len;

}

/* This routine is used to get the value of attribute name */
/* supplied as a parameter to this function. If successful, */
/* the routine returns a pointer to the SessionVariable */
/* object, else NULL. */

static SessionVariable *getVariable(text *t) {

  SessionVariable *current = session_variables;
  int len;

  len = VARSIZE(t) - VARHDRSZ;

  while (current != NULL) {
    
    if (current->keylen == len) {
      if (strncmp(current->key, VARDATA(t), len) == 0) {
	return current;
      }
    }
    current = current->child;

  }

  return NULL;

}

/* This routine is used to add a variable to the linked list. */
/* We always insert the variable at the head of the list. */
/* We must change the new child to reflect the addition. */

static void addVariable(text *key, text *value) {

  SessionVariable *v;

  /* Allocate a new variable */

  v = calloc(1, sizeof(SessionVariable));
  if (v == NULL) {
    elog(ERROR, "Unable to allocate new variable");
  }

  /* Set its key and value */

  setVariableAttribute(v, key, TRUE);
  setVariableAttribute(v, value, FALSE);

  /* Add it to the from of the list */

  if (session_variables != NULL) {
    session_variables->parent = v;
    v->child = session_variables;
    v->parent = NULL;
  }
  session_variables = v;

}

/* This routine is used to remove the variable from the list. */
/* The routine needs to check if the variable is the first or */
/* last in the list and take appropriate actions. */

static void removeVariable(SessionVariable *v) {
  
  if (v->parent == NULL) {
    session_variables = v->child;
    if (v->child != NULL) {
      v->child->parent = NULL;
    }
  } else {
    v->parent->child = v->child;
    if (v->child != NULL) {
      v->child->parent = v->parent;
    }
  }

  free(v->key);
  free(v->value);
  free(v);

} 

/* This function is used to set the value of a variable. */
/* It takes the name and value, possibly allocating a new map */
/* if necessary. The result of this function is -1 if the */
/* variable is removed, 0 if an existing variable is set, */
/* and 1 if the variable is added. */

PG_FUNCTION_INFO_V1(set_session_variable);

Datum set_session_variable(PG_FUNCTION_ARGS);
  
Datum set_session_variable(PG_FUNCTION_ARGS) {

  SessionVariable *current;
  text *key, *value;
  int32 added = 0;

  /* If the first argument is NULL, return NULL */

  if (PG_ARGISNULL(0)) {
    PG_RETURN_NULL();
  }
  key = (text *) PG_GETARG_TEXT_P(0);

  /* Attempt to find existing entry */

  current = getVariable(key);

  /* If the second argument is NULL, remove the maching entry, */
  /* else, either add or replace it. */

  if (PG_ARGISNULL(1)) {

    removeVariable(current);
    added = -1;

  } else {

    value = (text *) PG_GETARG_TEXT_P(1);

    if (current == NULL) {

      addVariable(key, value);
      added = 1;

    } else {

      setVariableAttribute(current, value, FALSE);
      added = 0;

    }

  }

  /* Return our added state */

  PG_RETURN_INT32(added);

}

/* This function is used to get the value of a variable. If the */
/* variable does not exist, NULL is returned. Otherwise the */
/* value of the variable is returned. */

PG_FUNCTION_INFO_V1(get_session_variable);

Datum get_session_variable(PG_FUNCTION_ARGS);

Datum get_session_variable(PG_FUNCTION_ARGS) {

  SessionVariable *current;
  text *key, *result;
  int32 len;

  key = (text *) PG_GETARG_TEXT_P(0);

  current = getVariable(key);
  
  if (current == NULL) {

    PG_RETURN_NULL();
  
  } else {

    len = strlen(current->value) + VARHDRSZ;
    result = (text *) palloc(len);
    VARATT_SIZEP(result) = len;
    memcpy(VARDATA(result), current->value, len - VARHDRSZ);

    PG_RETURN_TEXT_P(result);

  }

}

/* This function is used to reset all of the session variables. */
/* We just empty the list. This routine might need to be called */
/* when connection pooling is used, for example. The number of */
/* variables removed is returned as a result. */

PG_FUNCTION_INFO_V1(reset_session_variables);

Datum reset_session_variables(PG_FUNCTION_ARGS);

Datum reset_session_variables(PG_FUNCTION_ARGS) {

  SessionVariable *current, *child;
  int32 count = 0;
  
  current = session_variables;
  while (current != NULL) {
    child = current->child;
    removeVariable(current);
    count++;
    current = child;
  }

  PG_RETURN_INT32(count);

}


